/*
 * Copyright (C) 2018 Nick Downing <nick@ndcode.org>
 * SPDX-License-Identifier: MIT
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

let fs = require('fs')
let util = require('util')

let fs_stat = util.promisify(fs.stat)

let BuildCache = function(diag) {
  if (!this instanceof BuildCache)
    throw Error('BuildCache is a constructor')
  this.map = new Map()
  this.diag = diag || false
}

BuildCache.prototype.get = async function(key, build_func) {
  let result = this.map.get(key)
  if (result === undefined) {
    if (this.diag)
      console.log('building', key)
    result = {deps: [key], time: Date.now()}
    result.done = build_func(result)
    this.map.set(key, result)
    try {
      await result.done
    }
    catch (err) {
      delete result.done
      this.map.delete(key)
      throw err
    }
    delete result.done
  }
  else if (result.done === undefined) {
    if (this.diag)
      console.log('checking', key)
    result.done = (
      async () => {
        for (let i = 0; i < result.deps.length; ++i) {
          let stats
          try {
            stats = await fs_stat(result.deps[i])
          }
          catch (err) {
            if (!(err instanceof Error) || err.code !== 'ENOENT')
              throw err
            //stats = undefined
          }
          if (stats === undefined || stats.mtimeMs > result.time) {
            if (this.diag)
              console.log('rebuilding', key, 'reason', result.deps[i])
            result.deps = [key]
            result.time = Date.now()
            await build_func(result)
            break
          }
        }
      }
    )()
    try {
      await result.done
    }
    catch (err) {
      delete result.done
      this.map.delete(key)
      throw err
    }
    delete result.done
  }
  else
    await result.done
  return result.value
}

module.exports = BuildCache
